/****************************************************************

include/cons.h
Console declarations for AWOS

See the LICENSE file included with this source for copyright
information.

****************************************************************/

#ifndef ____CONS_H_
#define ____CONS_H_

#include <sys/types.h>
#include <stdint.h>
#include <stddef.h>

/** number of TTYs are defined here **/
#define NUM_TTYS 8

/** is being displayed **/
#define CONS_ACTIVE 0x00000001
/** can handle text routines **/
#define CONS_TEXTMODE 0x00000002
/** can handle graphics routines **/
#define CONS_GFXMODE 0x00000004
/** contents are persistent **/
#define CONS_PERSIST 0x00000008
/** GFX: transparency **/
#define CONS_TRANS 0x00000010

// It's GM's fault this thing is so damn complex
typedef struct console
{
	unsigned char *mem; /// Pointer to current position in memory
	unsigned char *ttymem; /// Pointer to TTY storage root
	unsigned char *vrammem; /// Pointer to VRAM storage root
	unsigned char *currmem; /// Pointer to current root
	ssize_t memlen; /// memory length
	int tx, ty, tw, th; /// TEXT: current x, y, width, height
	ssize_t tline; /// TEXT: line memory length
	int fw, fh; /// TEXT: Font width, height
	int gx, gy, gw, gh; /// GFX: current x, y, width, height
	ssize_t gline; /// GFX: line memory length
	uint32_t color, bcolor; /// color
	uint32_t flags; /// flags
	int tablen; /// tab length
	char *font; /// GFX: font
	char bpp; /// GFX: bpp
	uint32_t mode; /// GFX: mode
	char modetype; /// GFX: mode type
	int ww, wh, ws, wc; /// GFX: window width/height/size/count
	int pitch; /// GFX: pitch
	int i1, i2, i3, i4, i5; /// bytes you can play with
	void *v1, *v2, *v3, *v4, *v5; /// pointers you can play with
	struct cons_class *driver; /// driver functions
} console_t;

struct cons_class
{
	/** called on request **/
	void (*pop)(console_t *); // tty
	/** called on release **/
	void (*push)(console_t *); // tty
	/** TEXT/GFX: seek to a position in memory **/
	void (*tseek)(console_t *, int, int); // tty, x, y
	void (*gseek)(console_t *, int, int); // tty, x, y
	/** GFX: pixelwise routines **/
	void (*gincx)(console_t *); // tty
	void (*gincy)(console_t *); // tty
	void (*gdecx)(console_t *); // tty
	void (*gdecy)(console_t *); // tty
	/** TEXT: textwise routines **/
	void (*tincx)(console_t *); // tty
	void (*tincy)(console_t *); // tty
	void (*tdecx)(console_t *); // tty
	void (*tdecy)(console_t *); // tty
	/** TEXT: scroll one row down **/
	void (*vscroll)(console_t *); // tty
	/** TEXT: put a character onto the screen **/
	void (*putch)(console_t *, char); // tty
	/** GFX: plot a pixel at the current position **/
	void (*plot)(console_t *, int); // tty, colour
	/** TEXT/GFX: clear screen **/
	void (*cls)(console_t *); // tty
};

struct vesa_info
{
	uint16_t ax; // Feedback: return value
	
	uint8_t vesa_sig[4]; // "VESA"
	uint16_t vesa_ver;
	uint16_t oemstr_ptr, oemstr_seg;
	uint8_t capable;
#		define VESA_CAP_DAC8 0x0001
#		define VESA_CAP_VGA 0x0002
#		define VESA_CAP_XRAMDAC 0x0004
#		define VESA_CAP_STEREO 0x0008
#		define VESA_CAP_EXTSTEREO 0x0010
	uint16_t mode_ptr, mode_seg;
	uint16_t totalmemory; // 64KB blocks
	
	// VBE 2.0
	uint16_t oemrev;
	uint16_t oemvend_ptr, oemvend_seg;
	uint16_t oemprod_ptr, oemprod_seg;
	uint16_t oemprev_ptr, oemprev_seg;
	
	uint8_t reserved[222];
	uint8_t oemdata[256];
} __attribute__((__packed__));

struct vesa_mode_info
{
	// Feedback values from function
	uint16_t ax; // Status
	uint16_t cx; // Mode
	
	// VBE 1.0
	uint16_t mode_attr;
#		define VESA_MODEATTR_SUPPORTED 0x0001
#		define VESA_MODEATTR_BIOSOUTPUT 0x0004
#		define VESA_MODEATTR_COLOR 0x0008
#		define VESA_MODEATTR_GRAPHICS 0x0010
#		define VESA_MODEATTR_NONVGA 0x0020
#		define VESA_MODEATTR_NONBANK 0x0040
#		define VESA_MODEATTR_LINEAR 0x0080
#		define VESA_MODEATTR_DOUBLESCAN 0x0100
#		define VESA_MODEATTR_INTERLACE 0x0200
#		define VESA_MODEATTR_HW3BUF 0x0400
#		define VESA_MODEATTR_HWSTEREO 0x0800
#		define VESA_MODEATTR_DUALDISP 0x1000

	uint8_t wa_attr, wb_attr;
#		define VESA_WATTR_SUPPORTED 0x0001
#		define VESA_WATTR_READ 0x0002
#		define VESA_WATTR_WRITE 0x0003
	uint16_t w_gran, w_size;
	uint16_t wa_seg, wb_seg;
	uint16_t w_func_ptr, w_func_seg;
	uint16_t bytes_scanline;
	
	// VBE 1.2 (optional in VBE 1.0)
	uint16_t w, h;
	uint8_t char_w, char_h;
	uint8_t planes, bpp, banks, mem_model;
	uint8_t banksize; // in KB
	uint8_t images, rsv_1;
	
	// Direct Colour
	uint8_t r_size, r_pos;
	uint8_t g_size, g_pos;
	uint8_t b_size, b_pos;
	uint8_t rsv_size, rsv_pos;
	uint8_t dc_modeinfo;
	
	// VBE 2.0
	uint32_t phys_base_ptr;
	uint8_t reserved_vbe20[2];
	
	// VBE 3.0
	uint16_t lin_bytes_scanline;
	uint8_t bnk_imagepages;
	uint8_t lin_imagepages;
	uint8_t lin_r_size, lin_r_pos;
	uint8_t lin_g_size, lin_g_pos;
	uint8_t lin_b_size, lin_b_pos;
	uint8_t lin_rsv_size, lin_rsv_pos;
	uint32_t max_pixel_clock;
	uint8_t reserved[189];
} __attribute__((__packed__));

struct vesa_crtc_info
{
	uint16_t ax; // Feedback: return value
	uint16_t bx; // Mode
	
	uint16_t vtotal, vsync_s, vsync_e;
	uint16_t htotal, hsync_s, hsync_e;
	uint8_t flags;
#		define VESA_CRTC_DOUBLESCAN 0x01
#		define VESA_CRTC_INTERLACED 0x02
#		define VESA_CRTC_HPOLARITY 0x04
#		define VESA_CRTC_VPOLARITY 0x08
	uint32_t pixel_clock;
	uint16_t refresh_rate;
	uint8_t reserved[40];
} __attribute__((__packed__));

struct vesa_genreg
{
	uint16_t ax, bx, cx, dx;
	uint16_t si, di, bp, es;
};

#define TTY_MAKEDRV_TEXT(nam, pop, push, seek, tincx, tincy, tdecx, tdecy, vscroll, putch, cls) \
	void pop (console_t *); \
	void push (console_t *); \
	void seek (console_t *, int, int); \
	void tincx (console_t *); \
	void tincy (console_t *); \
	void tdecx (console_t *); \
	void tdecy (console_t *); \
	void vscroll (console_t *); \
	void putch (console_t *, char); \
	void cls (console_t *); \
	struct cons_class nam = { \
		pop, push, seek, NULL, NULL, NULL, NULL, NULL, tincx, tincy, tdecx, tdecy, vscroll, putch, NULL, cls \
	}

#define TTY_MAKEDRV_GFX(nam, pop, push, seek, gincx, gincy, gdecx, gdecy, plot, cls) \
	void pop (console_t *); \
	void push (console_t *); \
	void seek (console_t *, int, int); \
	void gincx (console_t *); \
	void gincy (console_t *); \
	void gdecx (console_t *); \
	void gdecy (console_t *); \
	void plot (console_t *, int); \
	void cls (console_t *); \
	struct cons_class nam = { \
		pop, push, NULL, seek, gincx, gincy, gdecx, gdecy, NULL, NULL, NULL, NULL, NULL, NULL, plot, cls \
	}

#define TTY_MAKEDRV_DUAL(nam, pop, push, tseek, gseek, gincx, gincy, gdecx, gdecy, tincx, tincy, tdecx, tdecy, vscroll, putch, plot, cls) \
	void pop (console_t *); \
	void push (console_t *); \
	void tseek (console_t *, int, int); \
	void gseek (console_t *, int, int); \
	void gincx (console_t *); \
	void gincy (console_t *); \
	void gdecx (console_t *); \
	void gdecy (console_t *); \
	void tincx (console_t *); \
	void tincy (console_t *); \
	void tdecx (console_t *); \
	void tdecy (console_t *); \
	void vscroll (console_t *); \
	void putch (console_t *, char); \
	void plot (console_t *, int); \
	void cls (console_t *); \
	struct cons_class nam = { \
		pop, push, tseek, gseek, gincx, gincy, gdecx, gdecy, tincx, tincy, tdecx, tdecy, vscroll, putch, plot, cls \
	}


/** This is where console memory starts. **/
extern unsigned char *con_start;

extern console_t ttys[NUM_TTYS];
extern console_t *currtty;

extern struct cons_class *cons_currclass;
extern struct cons_class cons_vga_80x25;

struct video_moder {
	console_t *(*fn)(int, int, int);
	struct video_moder *n;
};

extern char g_8x8_font[256*8*8/8];

// consinit.c
void ConInit(void);

// putchar.c
void putchar(const unsigned char);

// vgactl.c
void VgaUpdateCursor();
void VgaClearScreen();
console_t *VgaSetMode(int, int, int);
extern struct video_moder *vga_modelist,*vga_modelist_end;
void VgaSetConsole(console_t *tty);
//void VgaAddMode(console_t *(*)(int, int, int));

// vgagfx.c
void VgaGfxMode_gfx_320x200x8();
void VgaGfxMode_text_80x25();
console_t *video_modeset_vga13(int, int, int);
console_t *video_modeset_text80x25(int, int, int);

// prim.c
void GfxPlot(console_t *, int, int, int);
void GfxRect(console_t *, int, int, int, int, int);
void GfxRectFill(console_t *, int, int, int, int, int);
void GfxHLine(console_t *, int, int, int, int);
void GfxVLine(console_t *, int, int, int, int);
void GfxLine(console_t *, int, int, int, int, int);
void GfxTriSolid(console_t *, int,int, int,int, int,int, int);

// vesa.c
void VgaVesaInit(void);

/** VGA Index Register -- Sets the register to read from VGA_REG_DATA. **/
#define VGA_REG_IDX	0x3D4
/** VGA Data Register **/
#define VGA_REG_DATA	0x3D5

/* registers */
/** The start address of the screen (hardware scrolling) relative to B8000. **/
#define VGA_REG_STARTHI		0x0C
#define VGA_REG_STARTLO		0x0D
/** The cursor position **/
#define VGA_REG_CURSORHI	0x0E
#define VGA_REG_CURSORLO	0x0F

#endif

