/****************************************************************

kernel/gdtidtisr.c
GDT, IDT and ISR setup/configuration

See LICENSE file included with this source for copyright
information.

*****************************************************************/

#include <kernel.h>
#include <portio.h>
#include <stdint.h>
#include <gidt.h>
#include <string.h>
#include <cpuspec.h>	/* struct regs */
#include "syscall.h"	/* syscall_fn */

#define s(_a) void _a ();
s(isr00)
s(isr01)
s(isr02)
s(isr03)
s(isr04)
s(isr05)
s(isr06)
s(isr07)
s(isr08)
s(isr09)
s(isr0A)
s(isr0B)
s(isr0C)
s(isr0D)
s(isr0E)
s(isr0F)
s(isr10)
s(isr11)
s(isr12)
s(isr13)
s(isr14)
s(isr15)
s(isr16)
s(isr17)
s(isr18)
s(isr19)
s(isr1A)
s(isr1B)
s(isr1C)
s(isr1D)
s(isr1E)
s(isr1F)
s(isr20)
s(isr21)
s(isr22)
s(isr23)
s(isr24)
s(isr25)
s(isr26)
s(isr27)
s(isr28)
s(isr29)
s(isr2A)
s(isr2B)
s(isr2C)
s(isr2D)
s(isr2E)
s(isr2F)
s(isr30)
s(isr31)
s(isr32)
s(isr33)
s(isr34)
s(isr35)
s(isr36)
s(isr37)
s(isr38)
s(isr39)
s(isr3A)
s(isr3B)
s(isr3C)
s(isr3D)
s(isr3E)
s(isr3F)

s(irq00)
s(irq01)
s(irq02)
s(irq03)
s(irq04)
s(irq05)
s(irq06)
s(irq07)
s(irq08)
s(irq09)
s(irq0A)
s(irq0B)
s(irq0C)
s(irq0D)
s(irq0E)
s(irq0F)

s(isr_null)

#undef s

struct gidt_gate_entry idt[256];
struct gidt_block_entry gdt[8192];
struct gidt_ptr gdt_ptr;
struct gidt_ptr idt_ptr;
struct gidt_ptr tr_ptr;

struct tss ktss;

void gdt_update(void);
void idt_update(void);
void tr_update(uint32_t);

typedef struct irq_info
{
	void(*handler)(void *r);
	char *driver_name;
} irq_info_t;

irq_info_t KiIrqInfo[16];

void KiIrqInstall(int irq, void(*handler_fn)(void *r), char *drv_name)
{
	KiIrqInfo[irq].handler = handler_fn;
	KiIrqInfo[irq].driver_name = drv_name;
	
	return;
};

void KiIrqUninstall(int irq)
{
	KiIrqInfo[irq].handler = 0;
	KiIrqInfo[irq].driver_name = "Not in use.";

	return;
};

/****************************************************************
KiIrqHandler() - handles IRQ interrupts
Dispatches to proper handler function
Input: struct regs *r - registers
Output: Nothing
****************************************************************/

char irq_for_grabs[16]; // A way of getting IRQ notices to stuff which doesn't install a handler :\ -GM

void KiIrqHandler(struct regs *r)
{
	void (*handler)(struct regs *r);
	
	handler = (void (*)(struct regs *))KiIrqInfo[r->int_no].handler;
	if(handler)
		handler(r);
	else
	{
		if((r->int_no > 0x00) && (r->int_no != 0x0A))
			dprintf("KiIrqHandler: recieved unhandled IRQ number %02X\n", r->int_no);
		irq_for_grabs[r->int_no]++;
	};

	if(r->int_no >= 0x08)
		outb(0xA0, 0x20);

	outb(0x20, 0x20);
	
	return;
};

void KiRemapIrqs()
{
	//remap IRQs to start at INT 40h
	outb(0x20, 0x11);
	outb(0xA0, 0x11);
	outb(0x21, 0x40);
	outb(0xA1, 0x48);
	outb(0x21, 0x04);
	outb(0xA1, 0x02);
	outb(0x21, 0x01);
	outb(0xA1, 0x01);
	outb(0x21, 0x00);
	outb(0xA1, 0x00);
	
	return;
};



void KiConfigureTimer(void)
{
	uint16_t a = 1193182 / 100;
	outb(0x43, 0x26); // channel 0, lobyte/hibyte, square-wave, binary
	outb(0x40, a & 0xFF);
	outb(0x40, a >> 8);
};

void gdt_addblock(int idx, void *loc, uint32_t limit, uint16_t attribs)
{
	gdt[idx].base_low = ((uint32_t)loc & 0xFFFF);
	gdt[idx].base_mid = (((uint32_t)loc >> 16) & 0xFF);
	gdt[idx].base_high = ((uint32_t)loc >> 24);
	gdt[idx].limit_low = (limit & 0xFFFF);
	gdt[idx].attribs = ((limit >> 16) & 0x000F) | (attribs & 0xFFF0);
	gdt[idx].attribs = ((gdt[idx].attribs >> 8) & 0xFF) | ((gdt[idx].attribs << 8) & 0xFF00);
};

void idt_addblock(int idx, uint16_t seg, void *loc, uint16_t attribs)
{
	idt[idx].segment = seg;
	idt[idx].base_low = ((uint32_t)loc & 0xFFFF);
	idt[idx].base_high = ((uint32_t)loc >> 16);
	idt[idx].attribs = attribs;
};

void idt_addblock_stdi(int idx, void *loc)
{
	idt_addblock(idx, 0x0008, loc, IDT_PRESENT | IDT_DPL0 | IDT_INT);
};

void KsTableSetup()
{
	int i;
	
	asm("cli");
	
	gdt_ptr.loc = (uint32_t)&gdt[0];
	gdt_ptr.len = sizeof(struct gidt_block_entry) * 6 - 1;
	
	gdt_addblock(1, 0x00000000, 0xFFFFF, GDT_PRESENT | GDT_DPL0 | GDT_GRAN | GDT_BIG | GDT_CODE | GDT_CODE_READ  );
	gdt_addblock(2, 0x00000000, 0xFFFFF, GDT_PRESENT | GDT_DPL0 | GDT_GRAN | GDT_BIG | GDT_DATA | GDT_DATA_WRITE );
	ktss.backlink = 0;
	ktss.cr3 = 0;
	ktss.ldt = 0;
	ktss.iomap = 0xFFFF;
	
	/*
	unsigned int backlink;
	unsigned int esp0, ss0;
	unsigned int esp1, ss1;
	unsigned int esp2, ss2;
	unsigned int cr3;
	unsigned int eip, eflags;
	unsigned int eax, ecx, edx, ebx;
	unsigned int esp, ebp, esi, edi;
	unsigned int es, cs, ss, ds, fs, gs;
	unsigned int ldt;
	unsigned short iomap;
	unsigned short t;
	*/
	gdt_addblock(3, 0x00000000, 0xFFFFF, GDT_PRESENT | GDT_DPL3 | GDT_GRAN | GDT_BIG | GDT_CODE | GDT_CODE_READ );
	gdt_addblock(4, 0x00000000, 0xFFFFF, GDT_PRESENT | GDT_DPL3 | GDT_GRAN | GDT_BIG | GDT_DATA | GDT_DATA_WRITE);
	gdt_addblock(5, &ktss, sizeof(struct tss), GDT_PRESENT | GDT_DPL3 | GDT_TSS32);
	gdt_update();
	
	tr_update(0x0028);
	
	KiRemapIrqs();
	
	idt_ptr.loc = (uint32_t)&idt[0];
	idt_ptr.len = sizeof(struct gidt_gate_entry) * 256 - 1;
	
	idt_addblock_stdi(0x00, isr00);
	idt_addblock_stdi(0x01, isr01);
	idt_addblock_stdi(0x02, isr02);
	idt_addblock_stdi(0x03, isr03);
	idt_addblock_stdi(0x04, isr04);
	idt_addblock_stdi(0x05, isr05);
	idt_addblock_stdi(0x06, isr06);
	idt_addblock_stdi(0x07, isr07);
	idt_addblock_stdi(0x08, isr08);
	idt_addblock_stdi(0x09, isr09);
	idt_addblock_stdi(0x0A, isr0A);
	idt_addblock_stdi(0x0B, isr0B);
	idt_addblock_stdi(0x0C, isr0C);
	idt_addblock_stdi(0x0D, isr0D);
	idt_addblock_stdi(0x0E, isr0E);
	idt_addblock_stdi(0x0F, isr0F);
	
	idt_addblock_stdi(0x10, isr10);
	idt_addblock_stdi(0x11, isr11);
	idt_addblock_stdi(0x12, isr12);
	idt_addblock_stdi(0x13, isr13);
	idt_addblock_stdi(0x14, isr14);
	idt_addblock_stdi(0x15, isr15);
	idt_addblock_stdi(0x16, isr16);
	idt_addblock_stdi(0x17, isr17);
	idt_addblock_stdi(0x18, isr18);
	idt_addblock_stdi(0x19, isr19);
	idt_addblock_stdi(0x1A, isr1A);
	idt_addblock_stdi(0x1B, isr1B);
	idt_addblock_stdi(0x1C, isr1C);
	idt_addblock_stdi(0x1D, isr1D);
	idt_addblock_stdi(0x1E, isr1E);
	idt_addblock_stdi(0x1F, isr1F);
	
	idt_addblock_stdi(0x20, isr20);
	idt_addblock_stdi(0x21, isr21);
	idt_addblock_stdi(0x22, isr22);
	idt_addblock_stdi(0x23, isr23);
	idt_addblock_stdi(0x24, isr24);
	idt_addblock_stdi(0x25, isr25);
	idt_addblock_stdi(0x26, isr26);
	idt_addblock_stdi(0x27, isr27);
	idt_addblock_stdi(0x28, isr28);
	idt_addblock_stdi(0x29, isr29);
	idt_addblock_stdi(0x2A, isr2A);
	idt_addblock_stdi(0x2B, isr2B);
	idt_addblock_stdi(0x2C, isr2C);
	idt_addblock_stdi(0x2D, isr2D);
	idt_addblock_stdi(0x2E, isr2E);
	idt_addblock_stdi(0x2F, isr2F);
	
	idt_addblock_stdi(0x30, isr30);
	idt_addblock_stdi(0x31, isr31);
	idt_addblock_stdi(0x32, isr32);
	idt_addblock_stdi(0x33, isr33);
	idt_addblock_stdi(0x34, isr34);
	idt_addblock_stdi(0x35, isr35);
	idt_addblock_stdi(0x36, isr36);
	idt_addblock_stdi(0x37, isr37);
	idt_addblock_stdi(0x38, isr38);
	idt_addblock_stdi(0x39, isr39);
	idt_addblock_stdi(0x3A, isr3A);
	idt_addblock_stdi(0x3B, isr3B);
	idt_addblock_stdi(0x3C, isr3C);
	idt_addblock_stdi(0x3D, isr3D);
	idt_addblock_stdi(0x3E, isr3E);
	idt_addblock_stdi(0x3F, isr3F);
	
	idt_addblock_stdi(0x40, irq00);
	idt_addblock_stdi(0x41, irq01);
	idt_addblock_stdi(0x42, irq02);
	idt_addblock_stdi(0x43, irq03);
	idt_addblock_stdi(0x44, irq04);
	idt_addblock_stdi(0x45, irq05);
	idt_addblock_stdi(0x46, irq06);
	idt_addblock_stdi(0x47, irq07);
	idt_addblock_stdi(0x48, irq08);
	idt_addblock_stdi(0x49, irq09);
	idt_addblock_stdi(0x4A, irq0A);
	idt_addblock_stdi(0x4B, irq0B);
	idt_addblock_stdi(0x4C, irq0C);
	idt_addblock_stdi(0x4D, irq0D);
	idt_addblock_stdi(0x4E, irq0E);
	idt_addblock_stdi(0x4F, irq0F);
	
	for(i=0x50;i<0x100;i++){
		idt_addblock_stdi(i, isr_null);
	}
	
	idt_update();
	
	KiConfigureTimer();
};
