/***************************************************************
 *
 * arch/i386/sched.c
 * scheduler for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <sched.h>	/* scheduler defines */
#include <unistd.h>	/* BOOL, TRUE, FALSE */
#include <stdlib.h>	/* calloc(), free() */
#include <limits.h>	/* ATEXIT_MAX */
#include <stddef.h>	/* NULL */
#include <stdint.h>	/* types */
#include <macros.h>	/* assert, fixme */
#include <kernel.h>	/* dprintf(), panic() */
#include <v86.h>	/* KiV86Enter(), KiV86Return() */

#define STACK_SIZE	12288

#define CHECKFLAG(x)	(currtask->flags & x)

task_t idletask = NULL;
task_t currtask = NULL;
struct task dummytask;
extern struct tss ktss;

/* this is -1 because on first pre-increment it will be 0 */
static int32_t next_pid = -1;

static uint64_t currtime = 0;

extern void rand_tick(void);

/* functions that should NEVER be used outside this file, prototype here */
extern void __KiLoadSaneDescriptors(void);
BOOL __KiRegisterAtExitFunction(task_t, void (*)(void));
extern void tr_update(uint32_t);

void __KiUpdateTSS(struct regs *r)
{
	register uint32_t seg;
	if(!(r->cs & 3) && !(r->eflags & 0x20000))
		return;
	asm(
		"mov %%ss,%0"
		: "=a"(seg)
	);
	ktss.ss0 = ktss.ss1 = ktss.ss2 = seg;
	ktss.esp0 = ktss.esp1 = ktss.esp2 = (
		r->eflags & 0x20000
			? (uint32_t)(r+1)
			: (uint32_t)&(r->ves)
	);
	
	// FIXED: As it turns out, uncommenting this line causes a crash.
	// I have no idea why it crashes, all I know is that v86 in general
	// causes a ring 0 crash on my AMD Athlon XP 1900.
	// I have an nVidia GeForce 5200 AGP as my video card, but a VESA
	// mode query in itself should not cause a ring 0 fault. -GM
	//tr_update(0x0018);
	
	/*if(r->eflags & 0x20000)
		printf("[TSS=%08X]",ktss.esp0);*/
};

/* please note, I've taken out some of the more complex functionality.
 * I just want to start as simple as possible -- there may be an easier
 * way to do what we were doing than the way we were doing it --AW
 */
uint32_t *__KiGetNextTask(uint32_t *new_stack)
{
	task_t prevtask;

	__KiLoadSaneDescriptors();	// Fixes the V86 task switch bug. -AW
	
	assert(currtask != NULL);
	
	currtime++;
	rand_tick();
	
	/* GCC better like it this time. */
	//currtask->currstack = (void *)(&new_stack);
	currtask->currstack = new_stack; /* GCC would probably like this one better. */
	
	prevtask = currtask;
	/* Simple round-robin.  no priorities. */
	do
	{
		currtask = currtask->next;
		//dprintf("sched: %i-%08X-%08X-%08X\n", currtask->pid, currtask, currtask->currstack, currtask->next);
		if(CHECKFLAG(TSK_KILLED))
		{
			task_t next_task;
			
			dprintf("sched: pid %d dying...\n", currtask->pid);
			next_task = currtask->prev->next = currtask->next;
			currtask->next->prev = currtask->prev;
			if(prevtask == currtask){
				prevtask = currtask->next;
			}
			dprintf("sched: pid %d dead\n", currtask->pid);
			free(currtask->stack);
			free(currtask);
			if(next_task == currtask){
				panic("ALL TASKS HAVE BEEN KILLED INCLUDING THE IDLE TASK");
			}
			currtask = next_task;
		};
		if(CHECKFLAG(TSK_SLEEPING))
		{
			if(currtask->wake <= currtime)
				currtask->flags &= ~TSK_SLEEPING;
		};

		if(CHECKFLAG(TSK_RUNNABLE)
			&& !(CHECKFLAG(TSK_KILLED))
			&& !(CHECKFLAG(TSK_SLEEPING)))
		{
			__KiUpdateTSS((void *)(currtask->currstack));
			return currtask->currstack;
		};
	} while(currtask != prevtask);
	
	// Idling.
	currtask = idletask;
	//dprintf("sched: idling\n");
	
	return currtask->currstack;
};

void *__KiGetStack(void);
task_t __KiCreateTaskEx(uint32_t eip, char *name, char flags, char *cmdline, uint32_t extrastack)
{
	task_t new_task;
	uint32_t *stack, *esp, *ebp;
	uint8_t i;
	
	if(!currtask)
	{
		/* If currtask is NULL, add the current thread to the task list. */
		dprintf("sched: Initialising\n");
		currtask = calloc(sizeof(struct task));
		currtask->stack = __KiGetStack();
		currtask->name = "System Idle Process";
		currtask->pid = ++next_pid;
		//currtask->flags = TSK_RUNNABLE;
		currtask->flags = 0;
		currtask->cmdline = "";
		currtask->next = currtask->prev = currtask;
		currtask->stack = __KiGetStack();
		idletask = currtask;
	};
	
	new_task = calloc(sizeof(struct task));
	new_task->stack = calloc(STACK_SIZE);
	new_task->name = name;
	new_task->pid = ++next_pid;
	new_task->flags = flags;
	new_task->cmdline = cmdline;

	stack = new_task->stack + STACK_SIZE - extrastack;
	/* This extra wad of crap is to get V86 and ring 3 working -GM */
	*--stack = 0x0800;  /* V86 GS */
	*--stack = 0x0800;  /* V86 FS */
	*--stack = 0x0800;  /* V86 DS */
	*--stack = 0x0800;  /* V86 ES */
	
	esp = stack;
	*--stack = 0x0023;  /* User SS */
	*--stack = (uint32_t) esp;   /* User ESP */
	
	*--stack = 0x000202;	/* EFLAGS - corrected -GM */
	
	if(flags & TSK_USER) *--stack = 0x1B;
	else *--stack = 0x08;	/* CS */

	*--stack = eip;		/* EIP */
	
	*--stack = 0x00000000; /* Error code and int number for stack consistency */
	*--stack = 0x00000000;
	
	/* POPA */
	for(i = 0; i < 5; i++) *--stack = 0x41574F53;
	esp = stack;
	*--stack = 0x00EB4EB4;
	ebp = stack;
	*--stack = 0x00E51E51;
	*--stack = 0x00ED1ED1;
	/* data segment registers */
	for(i = 0; i < 4; i++) *--stack = ((flags & TSK_USER) ? 0x23 : 0x10);

	*esp = (unsigned long)stack;
	*ebp = (unsigned long)stack;

	new_task->currstack = stack;

	/* You'd better hope it doesn't break now, GM. -AW */
	dprintf("sched: Task: pid=%i/%i (%08X) (currtask=%08X).\n"
		,new_task->pid
		,next_pid
		,new_task
		,currtask
			);
	new_task->prev = currtask;
	new_task->next = currtask->next;
	currtask->next->prev = new_task;
	currtask->next = new_task;
	
	dprintf("sched: Task chain = %08X-%08X-%08X\n", currtask->prev, currtask, currtask->next);
	
	return new_task;
};

task_t __KiCreateTask(uint32_t eip, char *name)
{
	return __KiCreateTaskEx(eip, name, TSK_RUNNABLE, "", 0);
};

void __KiCreateChildTask(uint32_t eip, char *name, task_t parent)
{
	/* XXX: need a better way to do this than last time */
	fixme(); (void)eip; (void)name; (void)parent;
	return;
};

void __KiKillTask(task_t task)
{
	/* make sure the request is sane */
	assert(task != NULL);
	assert(task != idletask);
	assert(task->pid >= 2);

	task->flags |= TSK_KILLED;
	task->flags &= ~TSK_RUNNABLE;

	return;
};

void __KiKillTaskByPid(uint8_t pid)
{
	task_t loop;
	
	if(pid < 2)
	{
		dprintf("sched: attempted murder on kernel.\n");
		return;
	};

	for(loop = currtask; loop->pid != pid; loop = loop->next);

	__KiKillTask(loop);
};

/* NOTE TO PORTERS: To align with POSIX standard, "code" MUST be of type int. */
void exit(int code)
{
	(void)code;
	__KiKillTask(currtask);
	currtime--;
	asm("int $0x40");
	/* process is dead -- no way to get here */
};

int atexit(void (*func)(void))
{
	if(!(__KiRegisterAtExitFunction(currtask, func))) return -1;
	else return 0;
};

BOOL __KiRegisterAtExitFunction(task_t task, void (*function)(void))
{
	if(!(task->atexit_info.functions_reged < ATEXIT_MAX)) return FALSE;
	
	task->atexit_info.function[task->atexit_info.functions_reged++] = function;
	return TRUE;
};

void yield(void)
{
	currtime--;
	//rand_tick();
	asm("int $0x40");
};

void csleep(uint32_t cyc) /* sleep, in PIT cycles, and you're not going to sleep for 7.4 years. */
{
	currtask->wake = currtime + cyc;
	currtask->flags |= TSK_SLEEPING;
	//rand_tick();
	yield();
};

uint64_t get_time(void)
{
	return currtime;
};

