/**************************************************************
 *
 * drivers/ata/AtaController.c
 * ATA initialisation routines / driver entry point
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <stdint.h>
#include <sys/types.h>
#include <unistd.h>
#include <portio.h>
#include <macros.h>
#include <kalloc.h>
#include <string.h>
#include "ata.h"

void yield(void);

/*! \fn uint8_t AtaReadRegisterBsyWait(AtaController *controller, uint8_t Register)
 * \brief Waits for an ATA controller to clear its BSY flag, and then returns the contents of Register.
 * \param controller The ATA controller to read from.
 * \param Register The ATA register number to read from.
 * \return The contents of the ATA register specified by Register.
 * \warning This function does not check the validity of AtaDisk.
 */
uint8_t AtaReadRegisterBsyWait(AtaController *controller, uint8_t Register)
{
	uint8_t status;
	int i;
	
	for(i = 0; i < 50; i++)
	{
		status = inb(controller->basePort + ATA_REG_STATUS);
		if((status & ATA_STATUS_BSY) == 1) yield();
		else break;
	};
	
	if((status & ATA_STATUS_BSY) == 1)
	{
		printf("ata: waited 50 clock cycles with no response. disk failure.\n");
		return 0xFF;
	};
	
	return inb(controller->basePort + Register);
};

/*! \fn void AtaWriteRegister(AtaController *controller, uint8_t Register, uint8_t Data)
 * \brief Writes data to the specified register on the specified ATA controller.
 * \param controller The ATA controller to write to.
 * \param Register The register number to write to.
 * \param Data The data to write.
 * \return Nothing.
 * \note The caller should have a lock on the disk before calling this function.
 * \warning This function does not check the validity of controller.
 */
void AtaWriteRegister(AtaController *controller, uint8_t Register, uint8_t Data)
{
	int i;
	
	/* I'm sort of cheating here.  all registers require that
	 * BSY=0 and DRQ=0 before writing.  so I read the status port
	 * using BsyWait to take care of half the battle for me */
	for(i = 0; i < 50; i++)
	{
		if((AtaReadRegisterBsyWait(controller, ATA_REG_STATUS) & ATA_STATUS_DRQ) == 1) yield();
		else break;
	};
	
	if((AtaReadRegisterBsyWait(controller, ATA_REG_STATUS) & ATA_STATUS_DRQ) == 1)
	{
		printf("ata: waited 50 cycles with no response. disk failure.\n");
		return;
	};
	
	outb(controller->basePort + Register, Data);
	
	return;
};

/*! \fn AtaDisk AtaObtainDisk(AtaController *controller, uint8_t disk)
 * \brief Attempts to obtain a lock on a specified disk on a specified controller.
 * \param controller The ATA controller that the specified disk is on.
 * \param disk The disk to obtain a lock for.
 * \returns The disk if the lock was obtained, NULL if there was already a lock.
 */
AtaDisk AtaObtainDisk(AtaController *controller, uint8_t disk)
{
	AtaDisk Disk = controller->Disks[disk];
	
	if(Disk->Type != DRVTYPE_DISK_ATA) return NULL;	/* can't lock a non-ATA disk */
	
	if(Disk->Present == FALSE) return NULL;			/* can't lock a disk that isn't there */

	if(Disk->Locked == TRUE) return NULL;			/* can't lock a disk already locked */
	
	Disk->Locked = TRUE;
	return Disk;
};

void AtaReleaseDisk(AtaDisk disk)
{
	if(disk->Type != DRVTYPE_DISK_ATA) return;

	disk->Locked = FALSE;
	return;
};

/*! \fn AtaPowerMode AtaGetPowerMode(AtaDisk disk)
 * \brief Gets the current power mode for the specified ATA disk.
 * \param disk The handle of the disk to get the power mode for.
 * \return -1 on error, otherwise the power status.
 */
uint8_t AtaGetPowerMode(AtaDisk disk)
{
	uint8_t status;
	
	if(disk->Type != DRVTYPE_DISK_ATA) return -1;	/* check type descriptor */
	
	//if(::IsLocked(Disk) == FALSE) return -1;	/* get lock */
	
	AtaWriteRegister(disk->parent, ATA_REG_COMMAND, 0x98);
	status = AtaReadRegisterBsyWait(disk->parent, ATA_REG_SECCOUNT);
	
	if(status == ATA_STATUS_ERR) return -1;
	
	switch(status)
	{
		case 0x00:	/* ACTIVE or sometimes IDLE */
			return AtaPowerMode_Active;
		case 0x80:	/* IDLE */
			return AtaPowerMode_Idle;
		case 0xFF:	/* STANDBY */
			return AtaPowerMode_Standby;
		default:
			return -1;
	};
};

void AtaUnregisterDriver(int code)
{
	(void)code;
	return;
};

void AtaIrqHandler(void *r)
{
	(void)r;
/*	if(this->ExpectingIRQ == FALSE)
	{
		printf("ata: got an IRQ I wasn't expecting\n");
	};
	this->ExpectingIRQ = FALSE;
	(void)AtaReadRegisterBsyWait(ATA_REG_STATUS);*/
	return;
};

char *AtaGetSector(AtaController *controller, uint8_t numSectors)
{
	char *data;
	char *temp;

	data = (char *)kmalloc((sizeof(char) * 512) * numSectors);
	temp = (char *)kmalloc(sizeof(char) * 512);
	
	int i;
	
	/* I'm sort of cheating here.  all registers require that
	 * BSY=0 and DRQ=0 before writing.  so I read the status port
	 * using BsyWait to take care of half the battle for me */
	for(i = 0; i < 50; i++)
	{
		if((AtaReadRegisterBsyWait(controller, ATA_REG_STATUS) & ATA_STATUS_DRQ) == 1) yield();
		else break;
	};
	
	if((AtaReadRegisterBsyWait(controller, ATA_REG_STATUS) & ATA_STATUS_DRQ) == 1)
	{
		printf("ata: never got DRQ -- can't get sector\n");
		kfree(temp);
		kfree(data);
		return NULL;
	};
	
	for(i = 0; i < numSectors; i++)
	{
		insw(controller->basePort + ATA_REG_DATA, 256, (unsigned short *)temp);
		swab(temp, (data + (i * 512)), 512);
	};
	
	kfree(temp);

	return data;
};

char *AtaRead(AtaDisk Disk, BOOL Retry, uint8_t NumSectors, uint32_t StartSector)
{
	int i;
	
	for(i = 0; i < 50; i++)
	{
		if((AtaReadRegisterBsyWait(Disk->parent, ATA_REG_STATUS) & ATA_STATUS_DRDY) == 1) yield();
		else break;
	};
	
	if((AtaReadRegisterBsyWait(Disk->parent, ATA_REG_STATUS) & ATA_STATUS_DRDY) == 1)
	{
		printf("ata: waited 50 cycles with no response. disk failure.\n");
		return NULL;
	};
	
	AtaWriteRegister(Disk->parent, ATA_REG_SECCOUNT, NumSectors);
	AtaWriteRegister(Disk->parent, ATA_REG_LBALOW, (uint8_t)(StartSector));
	AtaWriteRegister(Disk->parent, ATA_REG_LBAMID, (uint8_t)(StartSector>>8));
	AtaWriteRegister(Disk->parent, ATA_REG_LBAHIGH,(uint8_t)(StartSector>>16));
	AtaWriteRegister(Disk->parent, ATA_REG_DEVHEAD, (0xE0 | (Disk->DiskNumber <<4) | ((StartSector >> 24) & 0x0F)));
	AtaWriteRegister(Disk->parent, ATA_REG_COMMAND, (Retry ? 0x20 : 0x21));
	
	return AtaGetSector(Disk->parent, NumSectors);
};

ata_disk_ident_t AtaGetIdent(AtaDisk Disk)
{
	ata_disk_ident_t id;
	char *sector;
	
	id = (ata_disk_ident_t)kmalloc(sizeof(struct ata_disk_ident));
	memset(id, sizeof(id), 0);
	
	Disk->parent->ExpectingIRQ = TRUE;
	AtaWriteRegister(Disk->parent, ATA_REG_DEVHEAD, (0xE0 | (Disk->DiskNumber << 4)));
	AtaWriteRegister(Disk->parent, ATA_REG_COMMAND, 0xEC);
	
	sector = AtaGetSector(Disk->parent, 1);
	memcpy(id, sector, 512);
	kfree(sector);
	
	return id;
};

/*! \fn BOOL AtaIsValidController (uint16_t port)
 * \brief Detects whether there is an ATA controller at the specified port.
 * \param port The port to scan.
 * \return TRUE if there is a controller at the port, FALSE otherwise.
 */
BOOL AtaIsValidController(uint16_t port)
{
	outb(port + ATA_REG_SECCOUNT, 0x3E);
	//csleep(20);
	if(inb(port + ATA_REG_SECCOUNT) != 0x3E)
		return FALSE;
	else return TRUE;
};

AtaController *AtaControllerInit(uint8_t irq, uint16_t port)
{
	AtaController *controller;
	uint8_t statusMaster, statusSlave;
	
	/* detect if there is an ATA controller */
	if(!AtaIsValidController(port)) return NULL;
	
	/* fill driver structs */
	KiIrqInstall(irq, AtaIrqHandler, "Generic ATA controller");
	controller = kmalloc(sizeof(AtaController));
	controller->Disks[0] = kcalloc(sizeof(AtaDisk));
	controller->Disks[1] = kcalloc(sizeof(AtaDisk));
	controller->basePort = port;
	controller->irq = irq;
	
	AtaWriteRegister(controller, ATA_REG_DEVHEAD, 0x00); //drive 0
	AtaWriteRegister(controller, ATA_REG_COMMAND, 0x90); //EXECUTE DRIVE DIAGNOSTICS
	statusMaster = AtaReadRegisterBsyWait(controller, ATA_REG_ERROR);
	AtaWriteRegister(controller, ATA_REG_DEVHEAD, 0x10); //drive 1
	AtaWriteRegister(controller, ATA_REG_COMMAND, 0x90); //EXECUTE DRIVE DIAGNOSTICS
	statusSlave = AtaReadRegisterBsyWait(controller, ATA_REG_ERROR);
	
	printf("ata: primary disk at %04X: %u\n", port, statusMaster);
	printf("ata: secondary disk at %04X: %u\n", port, statusSlave);
	
	controller->Disks[0]->parent = controller;
	controller->Disks[0]->Type = DRVTYPE_DISK_ATA;
	controller->Disks[0]->DiskNumber = 0;
	controller->Disks[0]->Locked = FALSE;
	if(statusMaster == 1)
	{
		controller->Disks[0]->Present = TRUE;
		controller->Disks[0]->Ident = AtaGetIdent(controller->Disks[0]);
		controller->Disks[0]->Ident->ModelName[39] = '\0';
		printf("ata: primary disk \"%s\"\n", controller->Disks[0]->Ident->ModelName);
	}
	else
		controller->Disks[0]->Present = FALSE;
	
	controller->Disks[1]->parent = controller;
	controller->Disks[1]->Type = DRVTYPE_DISK_ATA;
	controller->Disks[1]->DiskNumber = 1;
	controller->Disks[1]->Locked = FALSE;
	if(statusSlave == 1)
	{
		controller->Disks[1]->Present = TRUE;
		controller->Disks[1]->Ident = AtaGetIdent(controller->Disks[1]);
		controller->Disks[1]->Ident->ModelName[39] = '\0';
		printf("ata: secondary disk: \"%s\"\n", controller->Disks[1]->Ident->ModelName);
	}
	else
		controller->Disks[1]->Present = FALSE;
		
	return controller;
};

void AtaControllerDestroy(AtaController *controller)
{
	kfree(controller->Disks[1]);
	kfree(controller->Disks[0]);
	KiIrqUninstall(controller->irq);
	kfree(controller);
	return;
};

