/***************************************************************
 *
 * drivers/dma/dma.c
 * DMA handling routines for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <kernel.h>
#include <portio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <mutex.h>

#include <sched.h>

#include <dma.h>

void *dma_pointers[8] = {
	NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL
};

void *dma_old_pointers[8] = {
	NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL
};

uint16_t dma_io_pageregs[8] = {
	0x87, 0x83, 0x81, 0x82,
	0x8f, 0x8b, 0x89, 0x8a
};

uint16_t dma_io_addrregs[8] = {
	0x00, 0x02, 0x04, 0x06,
	0xc0, 0xc4, 0xc8, 0xcc
	// TODO: clarify: osdev.org says this is different (the 16-bit values),
	// whereas another doc i have suggests this is correct.
};

uint16_t dma_io_countregs[8] = {
	0x01, 0x03, 0x05, 0x07,
	0xc2, 0xc6, 0xca, 0xce
	// TODO: see above.
};

void dma_mutex_lock(unsigned int channel, void *optr, void *ptr)
{
	if(channel >= 8)
		return;
	
	cli();
	while(dma_pointers[channel] != NULL)
	{
		sti();
		yield();
		cli();
	};
	
	dma_old_pointers[channel] = optr;
	dma_pointers[channel] = ptr;
	sti();
};

BOOL dma_mutex_attempt_lock(unsigned int channel, void *optr, void *ptr)
{
	if(channel >= 8)
		return FALSE;
	
	cli();
	if(dma_pointers[channel] != NULL)
	{
		sti();
		return FALSE;
	};
	
	dma_old_pointers[channel] = optr;
	dma_pointers[channel] = ptr;
	sti();
	return TRUE;
};

void dma_mutex_unlock(unsigned int channel)
{
	cli();
	dma_pointers[channel] = NULL;
	if(dma_old_pointers[channel] != NULL)
	{
		free(dma_old_pointers[channel]);
		dma_old_pointers[channel] = NULL;
	}
	sti();
};

// OK, here are the main functions.

// These mask and unmask channels.
void DmaMaskChannel(unsigned int channel)
{
	uint8_t b, p;
	if(channel >= 8)
		return;
	p = (channel >= 4 ? 0xDE : 0x0F);
	b = inb(p) | (1<<(channel & 3));
	outb(p, b);
};

void DmaUnmaskChannel(unsigned int channel)
{
	uint8_t b, p;
	if(channel >= 8)
		return;
	p = (channel >= 4 ? 0xDE : 0x0F);
	b = inb(p) & ~(1<<(channel & 3));
	outb(p, b);
};

// This configures a channel, pointer-wise.
BOOL DmaConfigureChannelPointers(unsigned int channel, uint32_t size)
{
	// Sanity checks
	if(channel >= 8)
		return FALSE;
	
	if(dma_pointers[channel] == NULL)
		return FALSE;
	
	if(size <= 0)
		return FALSE;
	
	uint32_t pval = (uint32_t)dma_pointers[channel];
	uint16_t addr = (uint16_t)(channel >= 4 ? pval>>1 : pval); // truncate it
	uint16_t count = (uint16_t)(channel >= 4 ? size>>1 : size) - 1; // the "- 1" is a DMA caveat to be careful of
	uint8_t page = (uint8_t)(pval>>16);
	uint8_t rsff = (channel >= 4 ? 0xD8 : 0x0C); // ReSet Flip-Flop
	
	// Now, for the actual port writes...
	// Address.
	outb(rsff, 0xFF);
	outb(dma_io_addrregs[channel], (uint8_t)addr);
	outb(dma_io_addrregs[channel], (uint8_t)(addr >> 8));
	// Count.
	outb(rsff, 0xFF);
	outb(dma_io_countregs[channel], (uint8_t)count);
	outb(dma_io_countregs[channel], (uint8_t)(count >> 8));
	// Page.
	outb(dma_io_pageregs[channel], page);
	
	// Yay.
	return TRUE;
};

// This configures a channel, mode-wise.
BOOL DmaConfigureChannelMode(unsigned int channel, uint8_t transmode)
{
	if(channel >= 8)
		return FALSE;
	
	// Set the transfer mode.
	outb((channel >= 4 ? 0xD6 : 0x0B), (transmode & 0xFC) | (channel & 0x03));
	
	return TRUE;
};


// This creates a channel.
void *DmaCreateChannel(unsigned int channel, uint32_t size)
{
	void *ptr,*optr;
	
	if(channel >= 8)
		return FALSE;
	
	uint32_t align = (channel >= 4 ? 128<<10 : 64<<10);
	uint32_t realsize = (size<<1)-1;
	
	// 64KB for 8-bit, 128KB for 16-bit. Exactly that.
	if(size > align || size <= 0)
		return NULL;
	
	optr = malloc(realsize);
	if(optr == NULL)
		return NULL;
	
	if((uint32_t)optr > 0xFFFFFF) // 16MB limit :O
	{
		free(optr);
		return NULL;
	};
	
	ptr = optr;
	
	// 16-bit channels need 2-byte alignments
	if(channel >= 4)
		if((uint32_t)ptr & 1)
			ptr++;
	
	if((~(align-1) & (uint32_t)ptr) != (~(align-1) & ((uint32_t)ptr + size)))
		ptr = (void *)((~(align-1) & (uint32_t)ptr));
	dma_mutex_lock(channel, optr, ptr);
	
	// Mask the channel.
	DmaMaskChannel(channel);
	
	return ptr;
};

// This destroys a channel.
void DmaDestroyChannel(unsigned int channel)
{
	dma_mutex_unlock(channel);
};

// This configures a channel.

void DmaInit(void)
{
	MxReleaseMutex(MUTEX_DMA);
	MxGetMutex(MUTEX_DMA);
	
	dprintf("dma: locking unusable channels\n");
	
	dma_mutex_lock(0, NULL, (void *)((int)NULL + 1));
	dma_mutex_lock(4, NULL, (void *)((int)NULL + 1));
	
	MxReleaseMutex(MUTEX_DMA);
	
	printf("dma: initialised\n");
};
