/***************************************************************
 *
 * drivers/pci/pciinit.c
 * PCI initialisation routines for AWOS
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include <kernel.h>
#include <portio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <pci.h>

// 1/31: Enable Bit
// 7/30-24: Reserved
// 8/23-16: Bus Number
// 5/15-11: Slot Number
// 3/10-8: Function Number
// 6/7-2: Register Number
// 2/1-0: 00

struct pcidev_ll *pciroot = NULL;

BOOL pci_scanall = TRUE;

// PciReadConf32 is used for reading PCI information via an offset.
uint32_t PciReadConf32(uint16_t p, int bus, int slot, int funct, int offset)
{
	outd(p, 0x80000000
		| ((bus & 0x7F)<<16)
		| ((slot & 0x1F)<<11)
		| ((funct & 0x07)<<8)
		| (offset & 0xFC));
	
	return ind((int16_t)(p+0x4));
}

// PciWriteConf32 is used for writing PCI information via an offset.
void PciWriteConf32(uint16_t p, int bus, int slot, int funct, int offset, uint32_t data)
{
	outd(p, 0x80000000
		| ((bus & 0x7F)<<16)
		| ((slot & 0x1F)<<11)
		| ((funct & 0x07)<<8)
		| (offset & 0xFC));
	
	outd((int16_t)(p + 0x4), data);
}

// These 3 functions select a device by a specific class.
// A needs an exact class.
// B needs an exact class and subclass.
// C needs an exact class, subclass, and "prog. i/f".
//
// If pcidev is NULL, start scanning from the start.
// Otherwise, start scanning just after pcidev.
struct pcidev_ll *PciFindByClassA(struct pcidev_ll *pcidev, uint8_t cls_a)
{
	if(pcidev == NULL)
		pcidev = pciroot;
	else
		pcidev = pcidev->n;
	
	if(pciroot == NULL)
	{
		dprintf("pci: cannot search by class without a root\n");
		return NULL;
	};
	
	for(; pcidev; pcidev=pcidev->n)
	{
		if((pcidev->dev).cls_a == cls_a)
			return pcidev;
	};
	return NULL;
}

struct pcidev_ll *PciFindByClassB(struct pcidev_ll *pcidev, uint8_t cls_a, uint8_t cls_b)
{
	for(;;)
	{
		pcidev = PciFindByClassA(pcidev, cls_a);
		if(pcidev == NULL)
			return NULL;
		if((pcidev->dev).cls_b == cls_b)
			return pcidev;
	};
}

struct pcidev_ll *PciFindByClassC(struct pcidev_ll *pcidev, uint8_t cls_a, uint8_t cls_b, uint8_t cls_c)
{
	for(;;)
	{
		pcidev = PciFindByClassB(pcidev, cls_a, cls_b);
		if(pcidev == NULL)
			return NULL;
		if((pcidev->dev).cls_c == cls_c)
			return pcidev;
	};
}

// PciFindByVendor finds a device by vendor ID,
// much like the above 3.
// PciFindByProduct uses the product as well.
struct pcidev_ll *PciFindByVendor(struct pcidev_ll *pcidev, uint16_t vendor_id)
{
	if(pcidev == NULL)
		pcidev = pciroot;
	else
		pcidev = pcidev->n;
	
	if(pciroot == NULL)
	{
		dprintf("pci: cannot search by vendor without a root\n");
		return NULL;
	};
	
	for(; pcidev; pcidev=pcidev->n)
	{
		if((pcidev->dev).vendor_id == vendor_id)
			return pcidev;
	};
	return NULL;
}

struct pcidev_ll *PciFindByProduct(struct pcidev_ll *pcidev, uint16_t vendor_id, uint16_t device_id)
{
	for(;;)
	{
		pcidev = PciFindByVendor(pcidev, vendor_id);
		if(pcidev == NULL)
			return NULL;
		if((pcidev->dev).device_id == device_id)
			return pcidev;
	};
}


// PciScan is used for detecting PCI devices.
// There is no code for a PCI-to-PCI bridge;
// however, if you are to make one, just use this function
// with the port you obtained via any PCI bus.
void PciScan(uint16_t port)
{
	struct pcidev_ll *pcidev;
	uint32_t b,s,f;
	uint32_t i;
	
	dprintf("pci: scanning port %04X range\n", port);
	
	// If we forget to do this, then all but the first entry of pciroot
	// would be clipped off and cause a memory leak.
	if(pciroot != NULL)
	{
		for(pcidev = pciroot; pcidev->n != NULL;)
			pcidev = pcidev->n;
	};
	
	// Check each bus.
	for(b=0 ; b<128 ; b++) {
		i = PciReadConf32(port,b,0,0,0); // Read vendor ID.
		// A vendor ID of 0xFFFF means that the device does not exist.
		if(((i & 0xFFFF) != 0xFFFF) && pci_scanall)
		{
			// Check each slot.
			for(s=0 ; s<32 ; s++)
			{
				if(s != 0)
					i = PciReadConf32(port,b,s,0,0);
				if(((i & 0xFFFF) != 0xFFFF) && pci_scanall)
				{
					// Check each function.
					for(f=0 ; f<8 ; f++)
					{
						if(f != 0)
							i = PciReadConf32(port,b,s,f,0);
						
						if((i & 0xFFFF) == 0xFFFF)
							break; // Drop out regardless of the pcifastscan flag.
						
						
						dprintf("pci: device %04X:%i:%i:%i = %04X:%04X", port, b,s,f, i & 0xFFFF, i>>16);
						// Spawn pciroot if we have to,
						// else append to the end of the chain.
						if(pciroot == NULL)
						{
							pcidev = pciroot = malloc(sizeof(struct pcidev_ll));
							pcidev->p = pcidev->n = NULL;
						}
						else
						{
							pcidev->n = malloc(sizeof(struct pcidev_ll));
							pcidev->n->p = pcidev;
							pcidev = pcidev->n;
							pcidev->n = NULL;
						}
						
						// Right, let's copy all the info across.
						*((uint32_t *)(pcidev)) = i;
						for(i=4;i<sizeof(struct pcidev_t);i+=4)
						{
							*(uint32_t *)((uint8_t *)pcidev + i) = PciReadConf32(port,b,s,f,i);
						};
						dprintf(":%02X (class=%02X:%02X:%02X)\n", (pcidev->dev).revision, (pcidev->dev).cls_a, (pcidev->dev).cls_b, (pcidev->dev).cls_c);
						
						// Save the info, then move along.
						pcidev->port = port;
						pcidev->b = b;
						pcidev->s = s;
						pcidev->f = f;
					};
				};
			};
		};
	};
}

// PciInit is called early in the boot process.
void PciInit(void)
{
	printf("pci: scanning main PCI bus\n");
	if(k_kcmdline_check("pcifastscan") != NULL)
	{
		pci_scanall = FALSE;
		printf("pci: performing fast scan\n");
	};
	
	// This is the main PCI bus.
	PciScan(0x0CF8);
	return;
}

