/***************************************************************
 *
 * net/ifaces/e1000.cxx
 * Intel E1000-based Network Interface Controller Driver for ADM
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include "e1000.h"		// Net_E1000IfDrv proto
extern "C"
{
	#include <pci.h>		// Pci... fns
	#include <kernel.h>		// dprintf
};

Net_E1000IfDrv::Net_E1000IfDrv()
{
	dprintf("Intel E1000 series network card driver version 0.0.1\n");
	pciDevicesSupported[0] = 0x100E;	// 82540EM-A  (Desktop)
	pciDevicesSupported[1] = 0x100F;	// 82545EM-A  (Copper)
	pciDevicesSupported[2] = 0x1010;	// 82546EB-A1 (Copper Dual Port)
	pciDevicesSupported[3] = 0x1011;	// 82545EM-A  (Fiber)
	pciDevicesSupported[4] = 0x1012;	// 82546EB-A1 (Fiber Dual Port)
	pciDevicesSupported[5] = 0x1013;	// 82541EI-x0 (Cooper (Intel typo?))
	pciDevicesSupported[6] = 0x1015;	// 82540EM-A  (Mobile)
	pciDevicesSupported[7] = 0x1016;	// 82540EP-A  (Mobile)
	pciDevicesSupported[8] = 0x1017;	// 82540EP-A  (Desktop)
	pciDevicesSupported[9] = 0x1018;	// 82541EI-B0 (Mobile)
	pciDevicesSupported[10] = 0x1019;	// 82547xx    (Copper)
	pciDevicesSupported[11] = 0x101A;	// 82547EI-B0 (Mobile)
	pciDevicesSupported[12] = 0x101D;	// 82546EB-A1 (Copper Quad Port)
	pciDevicesSupported[13] = 0x1026;	// 82545GM-B  (Copper)
	pciDevicesSupported[14] = 0x1027;	// 82545GM-B  (Fiber)
	pciDevicesSupported[15] = 0x1028;	// 82545GM-B  (SerDes)
	pciDevicesSupported[16] = 0x1076;	// 82541xx    (Cooper)
	pciDevicesSupported[17] = 0x1077;	// 82541GI-B1 (Mobile)
	pciDevicesSupported[18] = 0x1078;	// 82541ER-C0 (Cooper)
	pciDevicesSupported[19] = 0x1079;	// 82546GB-B0 (Copper Dual Port)
	pciDevicesSupported[20] = 0x107A;	// 82546GB-B0 (Fiber Dual Port)
	pciDevicesSupported[21] = 0x107B;	// 82546GB-B0 (SerDes Dual Port)
	pciDevicesSupported[22] = 0x1107;	// 82544EI-A4 (Copper)
	pciDevicesSupported[23] = 0x1112;	// 82544GC-A4 (Copper)

	this->name = "Intel E1000 series";
};

Net_E1000IfDrv::~Net_E1000IfDrv()
{
};

BOOL Net_E1000IfDrv::InitializeDriver()
{
	BOOL keepSearching = TRUE;

	dprintf("e1000: searching for card");
	pciDevice = NULL;

	// There must be a prettier way to loop...
	while(keepSearching)
	{
		dprintf(".");
		pciDevice = PciFindByClassB(pciDevice, 2, 0); // net : eth
		if(pciDevice == NULL) return FALSE;	// there is no hope

		if(pciDevice->dev.vendor_id != 0x8086) continue;

		for(int i = 0; i < DEVS_SUPP; i++)
		{
			if(pciDevice->dev.device_id == pciDevicesSupported[i])
			{
				keepSearching = FALSE;
				break;
			};
		};
	};

	// We should never get here unless pciDevice is a valid E1000.

	dprintf("found\n");
	/* TODO: memory / page stuff goes here */
	
	// Set memory start to BAR0, and chop off the first four bits
	mem_start = (uint32_t *)((pciDevice->dev.bar[0]) & ~0xF);
	dprintf("e1000: device control registers @ %p\n", mem_start);

	uint32_t *dev_ctl = mem_start;
	uint32_t *dev_stat = mem_start + (NET_E1000_REG_STATUS / 4);
	
	dprintf("e1000: dev_ctl = %d, dev_stat = %d\n", *dev_ctl, *dev_stat);
	*dev_ctl |= (uint32_t)(1<<6);	/* set link up, i.e. turn this sucker on! */

	// let's get the MAC addy
	uint32_t *eep_reg = (uint32_t *)(mem_start + (0x14 / 4));
	uint32_t temp;

	for(int i = 0; i < 3; i ++)
	{
		temp = (i << 8);		// get the nth byte
		*eep_reg |= (temp | 0x01);	// reg addr and set read_req bit
		
		while((*eep_reg & 0x10) == 0)
			yield();		// I hope thats right

		mac_addr[(i * 2)] = ((uint8_t)(*eep_reg >> 16));
		mac_addr[((i * 2)+1)] = ((uint8_t)(*eep_reg >> 24));
	};

	return TRUE;
};

BOOL Net_E1000IfDrv::IsDriverInUse()
{
	return FALSE;
};

BOOL Net_E1000IfDrv::DisposeDriver()
{
	if(IsDriverInUse()) return FALSE;
	delete this;
	return TRUE;
};

int Net_E1000IfDrv::PushData(int type, void *data)
{
	(void)type;(void)data;
	return 0;
};

int Net_E1000IfDrv::RetrData(int type, void *data)
{
	(void)type;(void)data;
	return 0;
};

uint32_t Net_E1000IfDrv::GetDeviceStatus(void)
{
	return *(mem_start + (NET_E1000_REG_STATUS / 4));
};

uint32_t Net_E1000IfDrv::GetDeviceFlags(void)
{
	uint32_t devFlags = 0;
	uint32_t status;

	status = this->GetDeviceStatus();
	
	// convert device status to hardware-independent status
	if(status & 0x01)	// FD : Full Duplex
		devFlags |= NET_IFFLAG_DUPLEX;
	
	if(status & 0x02)	// LU : Link Up
		devFlags |= NET_IFFLAG_LINKUP;
	
	if(status & 0x40)	// SPEED : Link Speed
		devFlags |= NET_IFFLAG_SPEED_100M;
	
	return devFlags;
	
};

BOOL Net_E1000IfDrv::IsFlagSet(uint32_t flag)
{
	switch(flag)
	{
		case NET_IFFLAG_LINKUP:
			return (this->GetDeviceStatus() & 0x02);
		case NET_IFFLAG_DUPLEX:
			return (this->GetDeviceStatus() & 0x01);
		case NET_IFFLAG_SPEED_100M:
			return (this->GetDeviceStatus() & 0x40);
		default:
			return FALSE;
	};
};

BOOL Net_E1000IfDrv::SetDeviceFlag(uint32_t flag)
{
	uint32_t *devCtl = mem_start;

	switch(flag)
	{
		case NET_IFFLAG_LINKUP:
			// okay, lets bring it up!
			*devCtl |= 0x40;	// bit 6, SLU : Set Link Up
			
			return TRUE;
		
		default:
			return FALSE;
	};
};

