/***************************************************************
 *
 * net/ifaces/rtl8139.cxx
 * Realtek 8139-based Network Interface Controller Driver for ADM
 *
 * See LICENSE file included with this source for copyright
 * information.
 *
 ***************************************************************/

#include "rtl8139.h"

extern "C"
{
	#include <pci.h>	// Pci... fns
	#include <kernel.h>	// dprintf
	#include <portio.h>	// [in/out][b/w/l]
	#include <stdlib.h>	// malloc

	#include "../inet/ip_header.h"

};

Net_Rtl8139IfDrv::Net_Rtl8139IfDrv()
{
	dprintf("rtl8139 PCI network card driver version 0.0.1\n");
	
	pciDevicesSupported[0].vendor  = 0x10EC;
	pciDevicesSupported[0].product = 0x8139; // Realtek 8139

	pciDevicesSupported[1].vendor  = 0x10EC;
	pciDevicesSupported[1].product = 0x8129; // Realtek 8129

	pciDevicesSupported[2].vendor  = 0x10EC;
	pciDevicesSupported[2].product = 0x8138; // Realtek 8139B CardBus

	pciDevicesSupported[3].vendor  = 0x1113;
	pciDevicesSupported[3].product = 0x1211; // SMC1211TX

	pciDevicesSupported[4].vendor  = 0x1186;
	pciDevicesSupported[4].product = 0x1300; // D-Link DFE-538TX

	pciDevicesSupported[5].vendor  = 0x1186;
	pciDevicesSupported[5].product = 0x1301; // D-Link DFE-530TX+

	pciDevicesSupported[6].vendor  = 0x1186;
	pciDevicesSupported[6].product = 0x1340; // D-Link DFE-690TXD

	pciDevicesSupported[7].vendor  = 0x018A;
	pciDevicesSupported[7].product = 0x0106; // LevelOne FPC-0106TX

	pciDevicesSupported[8].vendor  = 0x021B;
	pciDevicesSupported[8].product = 0x8193; // Compaq HNE-300

	pciDevicesSupported[9].vendor  = 0x13D1;
	pciDevicesSupported[9].product = 0xAB06; // Edimax EP-4103DL CardBus

	pciDevicesSupported[10].vendor = 0x02AC;
	pciDevicesSupported[10].product= 0x1012; // Siemens 1012v2 CardBus

	pciDevicesSupported[11].vendor = 0x1432;
	pciDevicesSupported[11].product= 0x9130; // Siemens 1020

	this->name = (char *) "Realtek RTL8139 network card";
};

Net_Rtl8139IfDrv::~Net_Rtl8139IfDrv()
{
};

void Net_Rtl8139IfDrv::SetRegisterPage(unsigned int page)
{
	char b;
	
	// Sanity check
	if(page >= 4)
		return;
	
	b = inb(basePort + 0x00);
	dprintf("ne2000: inb 1 = %02X\n", b);
	b &= 0x3F;
	b |= (page<<6) & 0xC0;
	outb(basePort + 0x00, b);
	b = inb(basePort + 0x00);
	dprintf("ne2000: inb 2 = %02X\n", b);
};

BOOL Net_Rtl8139IfDrv::InitializeDriver()
{
	BOOL keepSearching = TRUE;
	int i;
	
	dprintf("rtl8139: searching for card");
	pciDevice = NULL;
	
	while(keepSearching)
	{
		dprintf(".");
		pciDevice = PciFindByClassB(pciDevice, 2, 0);
		if(pciDevice == NULL)
			return FALSE;

		for(int i = 0; i < DEVS_SUPP; i++)
		{
			if(pciDevice->dev.device_id == pciDevicesSupported[i].product
			&& pciDevice->dev.vendor_id == pciDevicesSupported[i].vendor)
			{
				keepSearching = FALSE;
				break;
			};
		};
	};
	
	dprintf("found\n");
	
	// Get basePort / irq
	basePort = pciDevice->dev.bar[0] & ~0x3;
	irq = pciDevice->dev.irq_line;
	dprintf("rtl8139: portbase %04X, irq %i\n", basePort, irq);
	
	// This is dervived from Realtek's own example code.
	// Reset card
	outb(basePort + RTL8139_REG_CMD, RTL8139_CMD_RESET);

	if((inb(basePort + RTL8139_REG_CMD) & RTL8139_CMD_RESET) != 0)
	{
		printf("rtl8139: failed resetting card.\r\n");
		return FALSE;
	}

	// enable card / link up
	outb(basePort + RTL8139_REG_CMD, RTL8139_CMD_RXON | RTL8139_CMD_TXON);

	// set card config
	outd(basePort + RTL8139_REG_TXCFG, RTL8139_TXC_IFG0 | RTL8139_TXC_IFG1 | RTL8139_TXC_MXDMA1);
	outd(basePort + RTL8139_REG_RXCFG, RTL8139_RXC_RBLEN0 | RTL8139_RXC_RBLEN1 | RTL8139_RXC_AM | RTL8139_RXC_MXDMA1 | RTL8139_RXC_MXDMA2 | RTL8139_RXC_APM | RTL8139_RXC_AB);

	// get a recv buffer
	rxBuffer = malloc((1024 * 64) + 16);
	outd(basePort + RTL8139_REG_RXBUF, (int32_t)rxBuffer);
	
	for(i=0; i<6; i++)
	{
		mac_addr[i] = inb(basePort + i);
	};
	
	printf("rtl8139: MAC address = %02X:%02X:%02X:%02X:%02X:%02X\n"
		,mac_addr[0]
		,mac_addr[1]
		,mac_addr[2]
		,mac_addr[3]
		,mac_addr[4]
		,mac_addr[5]
			);
	
	return TRUE;
};

BOOL Net_Rtl8139IfDrv::IsDriverInUse()
{
	return FALSE;
};

BOOL Net_Rtl8139IfDrv::DisposeDriver()
{
	if(IsDriverInUse())
		return FALSE;
	
	free(rxBuffer);	// no leaking, please
	delete this;
	return TRUE;
};

int Net_Rtl8139IfDrv::PushData(int type, void *data)
{
	(void) type;
	(void) data;

	return 0;
};

int Net_Rtl8139IfDrv::RetrData(int type, void *data)
{
	(void) type;
	(void) data;

	return 0;
};

uint32_t Net_Rtl8139IfDrv::GetDeviceFlags(void)
{
	return 0;
};

BOOL Net_Rtl8139IfDrv::IsFlagSet(uint32_t flag)
{
	(void) flag;
	return FALSE;
};

BOOL Net_Rtl8139IfDrv::SetDeviceFlag(uint32_t flag)
{
	(void) flag;
	return FALSE;
};
